#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#define PACKED __attribute__ ((packed))
typedef struct {
	unsigned short PACKED Reserved;
	unsigned short PACKED ResType;
	unsigned short PACKED ResCount;
} NEWHEADER;

typedef struct {
	unsigned char  PACKED Width;
	unsigned char  PACKED Height;
	unsigned char  PACKED ColorCount;
	unsigned char  PACKED reserved;
	unsigned short PACKED Planes;
	unsigned short PACKED BitCount;
	unsigned long  PACKED BytesInRes;
	unsigned long  PACKED FileOffset;
} RESDIR;

typedef struct {
	unsigned short PACKED bfType;
	unsigned long  PACKED bfSize;
	unsigned short PACKED bfReserved1;
	unsigned short PACKED bfReserved2;
	unsigned long  PACKED bfOffBits;
} BITMAPFILEHEADER;

typedef struct {
	unsigned long  PACKED biSize;
	long           PACKED biWidth;
	long           PACKED biHeight;
	unsigned short PACKED biPlanes;
	unsigned short PACKED biBitCount;
	unsigned long  PACKED biCompression;
	unsigned long  PACKED biSizeImage;
	unsigned long  PACKED biXPelsPerMeter;
	unsigned long  PACKED biYPelsPerMeter;
	unsigned long  PACKED biClrUsed;
	unsigned long  PACKED biClrImportant;
} BITMAPINFOHEADER;



int main(int argc, char **argv)
{
	int fh, bmp, bytes;
	char *workbuf, *datastart, *workptr;
	int i, j, ik=0;
	NEWHEADER *newheader;
	RESDIR *resdir;
	BITMAPFILEHEADER bitmapfh, bitmapfhm;
	BITMAPINFOHEADER bitmapih, bitmapihm;

	if (argc < 3) {
		printf("output.ico { image.bmp mask.bmp }\n");
		return -1;
	}

	workbuf = (char *)malloc(32768);	// bigass icon
	if (!workbuf) {
		printf("Error allocating work buffer\n");
		return -1;
	}
	memset(workbuf, 0, 32768);

	if (argc%2) argc--;
	ik = (argc-2)/2;

	newheader = (NEWHEADER *)workbuf;
	resdir    = (RESDIR *)(workbuf + sizeof(NEWHEADER));
	workptr   = workbuf + sizeof(NEWHEADER) + (sizeof(RESDIR) * ik);

	newheader->Reserved = 0;
	newheader->ResType = 1;
	newheader->ResCount = ik;

	for (i=0; i<ik; i++) {
		bmp = open(argv[i*2+2], O_RDONLY|O_BINARY);
		if (bmp<0) {
			free(workbuf);
			printf("%s not found\n", argv[i*2+2]);
			return -1;
		}

		bytes = read(bmp, &bitmapfh, sizeof(BITMAPFILEHEADER));
		if (bitmapfh.bfType != ('B'|('M'<<8))) {
			free(workbuf);
			printf("%s not a Windows bitmap\n", argv[i*2+2]);
			return -1;
		}

		bytes = read(bmp, &bitmapih, sizeof(BITMAPINFOHEADER));
		if (bitmapih.biCompression != 0) {
			free(workbuf);
			printf("%s not an RGB bitmap\n", argv[i*2+2]);
			return -1;
		}
		bitmapih.biSizeImage =
		bitmapih.biXPelsPerMeter =
		bitmapih.biYPelsPerMeter =
		bitmapih.biClrUsed =
		bitmapih.biClrImportant = 0;

		resdir[i].Width = bitmapih.biWidth;
		resdir[i].Height = bitmapih.biHeight;
		switch (bitmapih.biBitCount) {
			case 1:
				resdir[i].ColorCount = 2;
				break;

			case 4:
				resdir[i].ColorCount = 16;
				break;

			default:
				resdir[i].ColorCount = 0;
				break;
		}
		resdir[i].reserved = 0;
		resdir[i].Planes = bitmapih.biPlanes;
		resdir[i].BitCount = bitmapih.biBitCount;

		resdir[i].FileOffset = workptr - workbuf;

		datastart = workptr;

		bitmapih.biHeight <<= 1;
		memcpy(workptr, &bitmapih, sizeof(BITMAPINFOHEADER));
		workptr += sizeof(BITMAPINFOHEADER);
		bitmapih.biHeight >>= 1;

		j = bitmapih.biWidth;	// align mask to 32 bits
		if (j & 31) j = ((j >> 5) + 1) << 5;
		switch (bitmapih.biBitCount) {
			case 1:
				workptr += read(bmp, workptr, 4*2);	// palette
				workptr += read(bmp, workptr, (j * bitmapih.biHeight) >> 3);	// bitmap
				break;

			case 4:
				workptr += read(bmp, workptr, 4*16);
				workptr += read(bmp, workptr, (bitmapih.biWidth * bitmapih.biHeight) >> 1);
				break;

			case 8:
				workptr += read(bmp, workptr, 4*256);
				workptr += read(bmp, workptr, (bitmapih.biWidth * bitmapih.biHeight));
				break;

			default:
				workptr += read(bmp, workptr, (bitmapih.biWidth * bitmapih.biHeight) * (bitmapih.biBitCount >> 3));
				break;
		}
		/*
		memset(workptr, 0x00, (j * bitmapih.biHeight) >> 3);	// mask
		workptr += ((j * bitmapih.biHeight) >> 3);
		*/

		close(bmp);

		bmp = open(argv[i*2+2+1], O_RDONLY|O_BINARY);
		if (bmp<0) {
			free(workbuf);
			printf("%s not found\n", argv[i*2+2+1]);
			return -1;
		}

		bytes = read(bmp, &bitmapfhm, sizeof(BITMAPFILEHEADER));
		if (bitmapfhm.bfType != ('B'|('M'<<8))) {
			free(workbuf);
			printf("%s not a Windows bitmap\n", argv[i*2+2+1]);
			return -1;
		}

		bytes = read(bmp, &bitmapihm, sizeof(BITMAPINFOHEADER));
		if (bitmapihm.biCompression != 0) {
			free(workbuf);
			printf("%s not an RGB bitmap\n", argv[i*2+2+1]);
			return -1;
		}

		if (bitmapihm.biWidth != bitmapih.biWidth || bitmapihm.biHeight != bitmapih.biHeight) {
			free(workbuf);
			printf("%s is not the same size as the bitmap it's masking\n", argv[i*2+2+1]);
			return -1;
		}
		
		if (bitmapihm.biBitCount != 1) {
			free(workbuf);
			printf("%s is not a monochrome bitmap\n", argv[i*2+2+1]);
			return -1;
		}

		lseek(bmp, 2*4, 1);
		workptr += read(bmp, workptr, (j * bitmapih.biHeight) >> 3);	// copy mask data
		close(bmp);

		resdir[i].BytesInRes = workptr - datastart;
	}

	fh = open(argv[1], O_WRONLY|O_TRUNC|O_CREAT|O_BINARY, S_IREAD|S_IWRITE);
	if (!fh) {
		free(workbuf);
		printf("Error creating %s\n", argv[1]);
		return -1;
	}

	write(fh, workbuf, workptr-workbuf);

	close(fh);
	free(workbuf);

	return 0;
}
